from pedantic.class_decorators \
    import pedantic_class, pedantic_class_require_docstring, timer_class, trace_class, for_all_methods

from pedantic.method_decorators import pedantic, pedantic_require_docstring, trace, timer, require_not_none, \
    require_not_empty_strings, needs_refactoring, dirty, unimplemented, deprecated, overrides, validate_args, \
    require_kwargs

from pedantic.custom_exceptions import NotImplementedException, TooDirtyException



from typing import Callable, Any

# local file imports
from pedantic.method_decorators import pedantic, pedantic_require_docstring, trace, timer


def for_all_methods(decorator: Callable) -> Callable:
    """
    From: https://stackoverflow.com/questions/6307761/how-to-decorate-all-functions-of-a-class-without-typing-it-over-and-over-for-eac/6307868#6307868
    Example:
    >>> @for_all_methods(pedantic)
    >>> class C(object):
    >>>     def m1(self): pass
    >>>     def m2(self, x): pass
    """
    def decorate(cls: Any) -> Any:
        for attr in cls.__dict__:
            attr_value = getattr(cls, attr)

            if callable(attr_value):
                # if 'is_class_decorator' in inspect.getfullargspec(decorator).annotations: DOESNT WORK HERE
                try:
                    setattr(cls, attr, decorator(attr_value, is_class_decorator=True))
                except TypeError:
                    setattr(cls, attr, decorator(attr_value))
            elif isinstance(attr_value, property):
                prop = attr_value
                wrapped_getter = decorator(prop.fget) if prop.fget is not None else None
                wrapped_setter = decorator(prop.fset) if prop.fset is not None else None
                wrapped_deleter = decorator(prop.fdel) if prop.fdel is not None else None
                new_prop = property(fget=wrapped_getter, fset=wrapped_setter, fdel=wrapped_deleter)
                setattr(cls, attr, new_prop)  # write back new property
        return cls
    return decorate


def pedantic_class(cls: Any) -> Callable:
    """Shortcut for @for_all_methods(pedantic) """
    return for_all_methods(decorator=pedantic)(cls=cls)


def pedantic_class_require_docstring(cls: Any) -> Callable:
    """Shortcut for @for_all_methods(pedantic_require_docstring) """
    return for_all_methods(decorator=pedantic_require_docstring)(cls=cls)


def trace_class(cls: Any) -> Callable:
    """Shortcut for @for_all_methods(trace) """
    return for_all_methods(decorator=trace)(cls=cls)


def timer_class(cls: Any) -> Callable:
    """Shortcut for @for_all_methods(timer) """
    return for_all_methods(decorator=timer)(cls=cls)



import functools
import inspect
from typing import Callable, Any, Tuple, Dict, Type, Union
from datetime import datetime
import warnings
import re

# third party
from docstring_parser import parse, Docstring

# local file imports
from pedantic.custom_exceptions import NotImplementedException, TooDirtyException
from pedantic.type_hint_parser import is_instance


# HELPER FUNCTIONS ###
def __is_instance_method(func: Callable) -> bool:
    spec = inspect.getfullargspec(func)
    return spec.args and spec.args[0] == 'self'


def __is_static_method(func: Callable[..., Any]) -> bool:
    return '@staticmethod' in inspect.getsource(func)


def __is_property_setter(func: Callable[..., Any]) -> bool:
    return f'@{func.__name__}.setter' in inspect.getsource(func)


def __uses_multiple_decorators(func: Callable[..., Any], max_allowed: int = 1) -> bool:
    return len(re.findall('@', inspect.getsource(func).split('def')[0])) > max_allowed


def __get_args_without_self(func: Callable, args: Tuple[Any, ...], max_allowed) -> Tuple[Any, ...]:
    if __is_instance_method(func) or __is_static_method(func) or __uses_multiple_decorators(func, max_allowed):
        return args[1:]  # self is always the first argument if present
    return args


def __require_kwargs(func: Callable, args: Tuple[Any, ...], max_allowed: int) -> None:
    args_without_self = __get_args_without_self(func=func, args=args, max_allowed=max_allowed)

    if __is_special_func(func=func) or __is_property_setter(func=func):
        return

    assert args_without_self == (), f'{__qual_name(func=func)} Use kwargs when you call function {func.__name__}. ' \
                                    f'Args: {args_without_self}'


def __is_special_func(func: Callable[..., Any]) -> bool:
    name = func.__name__
    if not name.startswith('__') or not name.endswith('__'):
        return False
    return name not in [f'__{s}__' for s in ['new', 'init', 'str', 'del', 'int', 'float', 'complex', 'oct', 'hex',
                                             'index', 'trunc', 'repr', 'unicode', 'hash', 'nonzero', 'dir', 'sizeof']]


def __qual_name(func: Callable[..., Any]) -> str:
    """uniform design for error messages"""
    return f'In function {func.__qualname__}:' + '\n'


def __is_value_matching_type_hint(value: Any, type_hint: Any, func: Callable[..., Any]) -> bool:
    """Wrapper for file "type_hint_parser.py"."""

    if type_hint is None:
        return value == type_hint
    assert type(type_hint) is not tuple, \
        f'{__qual_name(func=func)} Use "Tuple[]" instead of "{type_hint}" as type hint.'
    assert type_hint is not tuple, f'{__qual_name(func=func)} Use "Tuple[]" instead of "tuple" as type hint.'
    assert type_hint is not list, f'{__qual_name(func=func)} Use "List[]" instead of "list" as type hint.'
    assert type_hint is not dict, f'{__qual_name(func=func)} Use "Dict[]" instead of "dict" as type hint.'
    assert type_hint is not set, f'{__qual_name(func=func)} Use "Set[]" instead of "set" as type hint.'
    assert type_hint is not frozenset, \
        f'{__qual_name(func=func)} Use "FrozenSet[]" instead of "frozenset" as type hint.'

    try:
        return is_instance(value, type_hint)
    except AssertionError as ex:
        raise AssertionError(f'{__qual_name(func=func)} {ex}')
    except (AttributeError, Exception) as ex:
        raise AssertionError(f'{__qual_name(func=func)} An error occurred during type hint checking. '
                             f'Value: {value} Annotation: {type_hint} '
                             f'Mostly this is caused by an incorrect type annotation. '
                             f'Details: {ex} ')


def __get_parsed_docstring(func: Callable) -> Docstring:
    docstring = func.__doc__.strip() if func.__doc__ is not None else ''
    try:
        return parse(docstring)
    except (Exception, TypeError) as ex:
        raise AssertionError(f'{__qual_name(func=func)} Could not parse docstring. Please check syntax. Details: {ex}')


def __get_annotations(func: Callable) -> Dict[str, Any]:
    return inspect.getfullargspec(func).annotations


def __require_kwargs_and_type_checking(func: Callable,
                                       args: Tuple[Any, ...],
                                       kwargs: Dict[str, Any],
                                       annotations: Dict[str, Any],
                                       signature: inspect.Signature,
                                       max_allowed: int) -> Any:
    """ Passing annotations and signature here has the advantage, that it only needs be calculated once
    and not during every function call.
    """
    __require_kwargs(func=func, args=args, max_allowed=max_allowed)
    params = signature.parameters

    assert signature.return_annotation is not inspect.Signature.empty, \
        f'{__qual_name(func=func)} Their should be a type hint for the return type ' \
        f'(e.g. None if there is nothing returned).'

    i = 1 if __is_instance_method(func=func) else 0
    for key in params:
        param = params[key]
        expected_type = param.annotation

        if param.name == 'self':
            continue

        assert expected_type is not inspect.Signature.empty, \
            f'{__qual_name(func=func)} Parameter "{param.name}" should have a type hint.'

        if param.default is inspect.Signature.empty:
            if __is_special_func(func=func) or __is_property_setter(func=func):
                actual_value = args[i]
                i += 1
            else:
                assert key in kwargs, f'{__qual_name(func=func)} Parameter "{key}" is unfilled.'
                actual_value = kwargs[key]
        else:
            actual_value = kwargs[key] if key in kwargs else param.default

        if isinstance(expected_type, str):
            class_name = actual_value.__class__.__name__
            assert class_name == expected_type, \
                f'{__qual_name(func=func)} Type hint is incorrect. Expected: {expected_type} but was {class_name} instead.'
        else:
            assert __is_value_matching_type_hint(value=actual_value, type_hint=expected_type, func=func), \
                f'{__qual_name(func=func)} Type hint is incorrect: Passed Argument {key}={actual_value} ' \
                f'does not have type {expected_type}.'

    result = func(*args, **kwargs) if not __is_static_method(func=func) else func(**kwargs)
    expected_result_type = annotations['return']

    if isinstance(expected_result_type, str):
        assert result.__class__.__name__ == expected_result_type, \
            f'{__qual_name(func=func)} Type hint is incorrect: Expected: {expected_result_type} ' \
            f'but was {result.__class__.__name__} instead.'
    else:
        assert __is_value_matching_type_hint(value=result, type_hint=expected_result_type, func=func), \
            f'{__qual_name(func=func)} Return type is incorrect: Expected {expected_result_type} ' \
            f'but {result} was the return value which does not match.'
    return result


def __require_docstring_google_format(func: Callable, docstring: Docstring, annotations: Dict[str, Any]) -> None:
    """
        Of course, the params docstring and annotations are obsolete here, but passing saves performance here,
        because they are only calculated once. If we would calculate them here, it would be very expensive.
    """
    num_documented_args = len(docstring.params)
    num_taken_args = len([a for a in annotations if a != 'return'])
    assert num_documented_args == num_taken_args, \
        f'{__qual_name(func=func)} There are {num_documented_args} argument(s) documented, but ' \
        f'{num_taken_args} are actually taken.'

    if docstring.returns is None:
        assert 'return' not in annotations or annotations['return'] is None, \
            f'{__qual_name(func=func)} The return type {annotations["return"]} is not documented.'
    else:
        assert 'return' in annotations and annotations['return'] is not None, \
            f'{__qual_name(func=func)} The return type {docstring.returns.type_name} is documented but the function ' \
            f'does not return anything.'

    for annotation in annotations:
        expected_type_raw = annotations[annotation]
        if 'typing.' in str(expected_type_raw):
            expected_type = str(annotations[annotation]).replace('typing.', '')
        elif hasattr(expected_type_raw, '__name__'):
            expected_type = expected_type_raw.__name__
        elif isinstance(expected_type_raw, str):
            expected_type = expected_type_raw
        else:
            expected_type = None

        if annotation == 'return' and annotations[annotation] is not None:
            assert len(docstring.returns.args) == 2, \
                f'{__qual_name(func=func)} Parsing Error. Only Google style Python docstrings are supported.'

            actual_return_type: str = docstring.returns.args[1]
            assert actual_return_type == expected_type, \
                f'{__qual_name(func=func)} Documented type is incorrect: Annotation: {expected_type} ' \
                f'Documented: {actual_return_type}'
        elif annotation != 'return':  # parameters passed to function
            docstring_param = None
            for param in docstring.params:
                if param.arg_name == annotation:
                    docstring_param = param
            assert docstring_param is not None, f'{__qual_name(func=func)} Parameter {annotation} is not documented.'
            assert expected_type == docstring_param.type_name, \
                f'{__qual_name(func=func)} Documented type of parameter {annotation} is incorrect. ' \
                f'Expected {expected_type} but documented is {docstring_param.type_name}.'


def __raise_warning(msg: str, category: Type[Warning]):
    warnings.simplefilter('always', category)  # turn off filter
    warnings.warn(msg, category=category, stacklevel=2)
    warnings.simplefilter('default', category)  # reset filter


# DECORATORS ###
def overrides(interface_class: Any) -> Callable:
    """
        Example:
        >>> class Parent:
        >>>     def instance_method(self):
        >>>         pass
        >>> class Child(Parent):
        >>>     @overrides(Parent)
        >>>     def instance_method(self):
        >>>         print('hi')
    """
    def overrider(func: Callable) -> Callable:
        assert __is_instance_method(func) or __uses_multiple_decorators(func), \
            f'Function "{func.__name__}" is not an instance method!'
        assert (func.__name__ in dir(interface_class)), \
            f"Parent class {interface_class.__name__} does not have such a method '{func.__name__}'."
        return func
    return overrider


def timer(func: Callable) -> Callable:
    """
        Prints out how long the execution of the function takes in seconds.
        Example:
        >>> @timer
        >>> def some_calculation():
        >>>     return 42
    """
    @functools.wraps(func)
    def wrapper(*args, **kwargs) -> Any:
        start_time: datetime = datetime.now()
        value: Any = func(*args, **kwargs)
        end_time = datetime.now()
        run_time = end_time - start_time
        print("Timer: Finished function {} in {}.".format(repr(func.__name__), run_time))
        return value
    return wrapper


def count_calls(func: Callable) -> Callable:
    """
        Prints how often the method is called during program execution.
        Example:
        >>> @count_calls
        >>> def some_calculation():
        >>>     return 42
    """
    @functools.wraps(func)
    def wrapper(*args, **kwargs) -> Any:
        wrapper.num_calls += 1
        print(f"Count Calls: Call {wrapper.num_calls} of function {func.__name__!r} at {datetime.now()}.")
        return func(*args, **kwargs)

    wrapper.num_calls = 0
    return wrapper


def trace(func: Callable) -> Callable:
    """
       Prints the passed arguments and keyword arguments and return values on each function call.
       Example:
       >>> @trace
       >>> def some_calculation(a, b, c):
       >>>     return a + b + c
    """
    @functools.wraps(func)
    def wrapper(*args, **kwargs) -> Any:
        print(f'Trace: {datetime.now()} calling {func.__name__}()  with {args}, {kwargs}')
        original_result = func(*args, **kwargs)
        print(f'Trace: {datetime.now()} {func.__name__}() returned {original_result!r}')
        return original_result
    return wrapper


def deprecated(func: Callable) -> Callable:
    """This is a decorator which can be used to mark functions as deprecated. It will result in a warning being emitted
    when the function is used."""
    @functools.wraps(func)
    def new_func(*args, **kwargs):
        __raise_warning(msg="Call to deprecated function {}.".format(func.__name__), category=DeprecationWarning)
        return func(*args, **kwargs)
    return new_func


def needs_refactoring(func: Callable) -> Callable:
    """
        Of course, you refactor immediately if you see something ugly.
        However, if you don't have the time for a big refactoring use this decorator at least.
        A warning is printed everytime the decorated function is called.
    """
    @functools.wraps(func)
    def wrapper(*args, **kwargs) -> Any:
        __raise_warning(msg=f'Function "{func.__name__}" looks terrible and needs a refactoring!', category=UserWarning)
        return func(*args, **kwargs)
    return wrapper


def unimplemented(func: Callable) -> Callable:
    """
        For documentation purposes. Throw NotImplementedException if the function is called.
        Example:
        >>> @unimplemented
        >>> def some_calculation(a, b, c):
        >>>     pass
        >>> some_calculation(5, 4, 3)       # this will lead to an NotImplementedException
    """
    @functools.wraps(func)
    def wrapper(*args, **kwargs) -> Any:
        raise NotImplementedException(f'Function "{func.__name__}" is not implemented yet!')
    return wrapper


def dirty(func: Callable) -> Callable:
    """
        Prevents dirty code from beeing executed.
        Example:
        >>> @dirty
        >>> def some_calculation(a, b, c):
        >>>     return a + a + a + a + a
        >>> some_calculation(5, 4, 3)       # this will lead to an TooDirtyException
    """
    @functools.wraps(func)
    def wrapper(*args, **kwargs) -> Any:
        raise TooDirtyException(f'Function "{func.__name__}" is too dirty to be executed!')
    return wrapper


def require_kwargs(func: Callable, is_class_decorator: bool = False) -> Callable:
    """
        Checks that each passed argument is a keyword argument.
        Example:
        >>> @require_kwargs
        >>> def some_calculation(a, b, c):
        >>>     return a + b + c
        >>> some_calculation(5, 4, 3)       # this will lead to an AssertionError
        >>> some_calculation(a=5, b=4, c=3) # this is okay
    """
    @functools.wraps(func)
    def wrapper(*args, **kwargs) -> Any:
        __require_kwargs(func=func, args=args, max_allowed=0 if is_class_decorator else 1)
        return func(*args, **kwargs)
    return wrapper


def validate_args(validator: Callable[[Any], Union[bool, Tuple[bool, str]]],
                  is_class_decorator: bool = False) -> Callable:
    """
      Validates each passed argument with the given validator.
      Example:
      >>> @validate_args(lambda x: (x > 42, f'Each argument should be greater then 42, but it was {x}.'))
      >>> def some_calculation(a, b, c):
      >>>     return a + b + c
      >>> some_calculation(30, 40, 50)  # this leads to an error
      >>> some_calculation(43, 40, 50)  # this is okay
   """
    def outer(func: Callable) -> Callable:
        def validate(obj: Any) -> None:
            res = validator(obj)
            res, msg = res if type(res) is not bool else (res, 'Invalid arguments.')
            assert res, f'{__qual_name(func=func)} {msg}'

        @functools.wraps(func)
        def wrapper(*args, **kwargs) -> Any:
            args_without_self = __get_args_without_self(func=func,
                                                        args=args,
                                                        max_allowed=0 if is_class_decorator else 1)

            for arg in args_without_self:
                validate(arg)

            for kwarg in kwargs:
                validate(kwargs[kwarg])

            return func(*args, **kwargs)
        return wrapper
    return outer


def require_not_none(func: Callable, is_class_decorator: bool = False) -> Callable:
    """
      Checks that each passed argument is not None and raises AssertionError if there is some.
      Example:
      >>> @require_not_none
      >>> def some_calculation(a, b, c):
      >>>     return a + b + c
   """
    def validator(arg: Any) -> Tuple[bool, str]:
        return arg is not None, f'Argument of function "{func.__name__}" should not be None!'
    return validate_args(validator=validator, is_class_decorator=is_class_decorator)(func)


def require_not_empty_strings(func: Callable, is_class_decorator: bool = False) -> Callable:
    """
       Throw a ValueError if the arguments are None, not strings, or empty strings or contains only whitespaces.
       Example:
       >>> @require_not_empty_strings
       >>> def some_calculation(a, b, c):
       >>>     print(a, b, c)
       >>> some_calculation('Hi', '   ', 'you')    # this will lead to an ValueError
       >>> some_calculation('Hi', None, 'you')     # this will lead to an ValueError
       >>> some_calculation('Hi', 42, 'you')       # this will lead to an ValueError
   """
    def validator(arg: Any) -> Tuple[bool, str]:
        return arg is not None and type(arg) is str and len(arg.strip()) > 0, \
            f'Arguments of function "{func.__name__}" should be a not empty string! Got: {arg}'
    return validate_args(validator=validator, is_class_decorator=is_class_decorator)(func)


def pedantic(func: Callable, is_class_decorator: bool = False) -> Callable:
    """
       Checks the types and throw an AssertionError if a type is incorrect.
       This decorator reads the type hints and use them as contracts that will be checked.
       If the function misses type hints, it will raise an AssertionError.
       It also forces the usage of keyword arguments. Otherwise, type checking would be impossible.

       Example:
       >>> @pedantic
       >>> def some_calculation(a: int, b: float, c: str) -> bool:
       >>>     print(a, b, c)
       >>>     return float(a) == b
       >>> some_calculation(a=42.0, b=14.0, c='you')   # this will lead to an AssertionError
       >>> some_calculation(a=42, b=None, c='you')     # this will lead to an AssertionError
       >>> some_calculation(a=42, b=42, c='you')       # this will lead to an AssertionError
       >>> some_calculation(5, 4.0, 'hi')              # this will lead to an AssertionError
   """

    annotations = __get_annotations(func=func)
    docstring = __get_parsed_docstring(func=func)

    if len(__get_parsed_docstring(func=func).params) > 0:
        __require_docstring_google_format(func=func, annotations=annotations, docstring=docstring)

    @functools.wraps(func)
    def wrapper(*args, **kwargs) -> Any:
        return __require_kwargs_and_type_checking(func=func,
                                                  args=args,
                                                  kwargs=kwargs,
                                                  annotations=annotations,
                                                  signature=inspect.signature(func),
                                                  max_allowed=0 if is_class_decorator else 1
                                                  )
    return wrapper


def pedantic_require_docstring(func: Callable, is_class_decorator: bool = False) -> Callable:
    """It's like pedantic, but now it forces you to write docstrings (Google format)."""

    annotations = __get_annotations(func=func)
    docstring = __get_parsed_docstring(func=func)

    def wrapper(*args, **kwargs) -> Any:
        __require_docstring_google_format(func=func, annotations=annotations, docstring=docstring)
        res = __require_kwargs_and_type_checking(func=func,
                                                 args=args,
                                                 kwargs=kwargs,
                                                 annotations=annotations,
                                                 signature=inspect.signature(func),
                                                 max_allowed=0 if is_class_decorator else 1)
        return res
    return wrapper



import unittest
import sys
import os
sys.path.append(os.getcwd())

# local file imports
from pedantic.unit_tests.tests_require_docstrings_google_format import TestRequireDocstringGoogleFormat
from pedantic.unit_tests.tests_require_kwargs import TestRequireKwargs
from pedantic.unit_tests.tests_class_decorators import TestClassDecorators
from pedantic.unit_tests.tests_require_kwargs_and_type_check import TestDecoratorRequireKwargsAndTypeCheck
from pedantic.unit_tests.tests_small_method_decorators import TestSmallDecoratorMethods
from pedantic.unit_tests.tests_combination_of_decorators import TestCombinationOfDecorators

if __name__ == '__main__':
    unittest.main()



import unittest
from abc import ABC, abstractmethod

# local file imports
from pedantic.class_decorators import pedantic_class, for_all_methods
from pedantic.method_decorators import overrides, validate_args, pedantic


class TestCombinationOfDecorators(unittest.TestCase):

    def test_pedantic_overrides(self):
        class MyClass(ABC):
            @pedantic
            @abstractmethod
            def op(self, a: int) -> None:
                pass

        class Child(MyClass):
            a = 0

            @pedantic
            @overrides(MyClass)
            def op(self, a: int) -> None:
                self.a = a

        c = Child()
        c.op(a=42)

    def test_pedantic_validate_args_1(self):
        @validate_args(lambda x: (x > 0, f'Argument should be greater then 0, but it was {x}.'))
        @pedantic
        def some_calculation(x: int) -> int:
            return x

        some_calculation(x=42)
        with self.assertRaises(expected_exception=AssertionError):
            some_calculation(x=0)
        with self.assertRaises(expected_exception=AssertionError):
            some_calculation(x=-42)
        with self.assertRaises(expected_exception=AssertionError):
            some_calculation(x=1.0)

    def test_pedantic_validate_args_2(self):
        """Different order decorators"""
        @pedantic
        @validate_args(lambda x: (x > 0, f'Argument should be greater then 0, but it was {x}.'))
        def some_calculation(x: int) -> int:
            return x

        some_calculation(x=42)
        with self.assertRaises(expected_exception=AssertionError):
            some_calculation(x=0)
        with self.assertRaises(expected_exception=AssertionError):
            some_calculation(x=-42)
        with self.assertRaises(expected_exception=AssertionError):
            some_calculation(x=1.0)

    def test_pedantic_validate_args_3(self):
        class MyClass:
            @pedantic
            @validate_args(lambda x: x > 0)
            def some_calculation(self, x: int) -> int:
                return x

        m = MyClass()
        m.some_calculation(x=42)
        with self.assertRaises(expected_exception=AssertionError):
            m.some_calculation(x=0)
        with self.assertRaises(expected_exception=AssertionError):
            m.some_calculation(x=-42)
        with self.assertRaises(expected_exception=AssertionError):
            m.some_calculation(x=1.0)

    def test_pedantic_validate_args_4(self):
        @pedantic_class
        class MyClass:
            @validate_args(lambda x: x > 0)
            def some_calculation(self, x: int) -> int:
                return x

        m = MyClass()
        m.some_calculation(x=42)
        with self.assertRaises(expected_exception=AssertionError):
            m.some_calculation(x=0)
        with self.assertRaises(expected_exception=AssertionError):
            m.some_calculation(x=-42)
        with self.assertRaises(expected_exception=AssertionError):
            m.some_calculation(x=1.0)

    def test_pedantic_validate_args_5(self):
        @pedantic
        @validate_args(lambda x: x > 0)
        def some_calculation(x: int) -> int:
            return x

        some_calculation(x=42)
        with self.assertRaises(expected_exception=AssertionError):
            some_calculation(x=0)
        with self.assertRaises(expected_exception=AssertionError):
            some_calculation(x=-42)
        with self.assertRaises(expected_exception=AssertionError):
            some_calculation(42)

    def test_pedantic_validate_args_6(self):
        class MyClass:
            @pedantic
            @validate_args(lambda x: x > 0)
            def some_calculation(self, x: int) -> int:
                return x

        m = MyClass()
        m.some_calculation(x=42)
        with self.assertRaises(expected_exception=AssertionError):
            m.some_calculation(x=0)
        with self.assertRaises(expected_exception=AssertionError):
            m.some_calculation(x=-42)
        with self.assertRaises(expected_exception=AssertionError):
            m.some_calculation(42)

    def test_pedantic_class_static_method_1(self):
        @pedantic_class
        class MyClass:
            @staticmethod
            def some_calculation(x: int) -> int:
                return x

        m = MyClass()
        m.some_calculation(x=42)
        MyClass.some_calculation(x=45)

    def test_pedantic_class_static_method_2(self):
        """Never do this, but it works"""
        @for_all_methods(staticmethod)
        @pedantic_class
        class MyClass:
            def some_calculation(x: int) -> int:
                return x

        m = MyClass()
        m.some_calculation(x=42)
        with self.assertRaises(expected_exception=AssertionError):
            m.some_calculation(x=42.0)
        MyClass.some_calculation(x=45)
        with self.assertRaises(expected_exception=AssertionError):
            MyClass.some_calculation(x=45.0)

    def test_pedantic_static_method_1(self):
        class MyClass:
            @staticmethod
            @pedantic
            def some_calculation(x: int) -> int:
                return x

        m = MyClass()
        m.some_calculation(x=42)
        MyClass.some_calculation(x=45)


if __name__ == '__main__':
    t = TestCombinationOfDecorators()
    t.test_pedantic_class_static_method_1()



import unittest
import warnings

# local file imports
from pedantic.custom_exceptions import TooDirtyException, NotImplementedException
from pedantic.method_decorators import overrides, deprecated, needs_refactoring, dirty, timer, count_calls, \
    unimplemented, validate_args, require_not_none, require_not_empty_strings


class TestSmallDecoratorMethods(unittest.TestCase):

    def test_overrides(self):
        """Problem here: parent has no such method"""
        class A:
            pass

        with self.assertRaises(expected_exception=AssertionError):
            class B(A):
                @overrides(A)
                def operation(self):
                    return 42

    def test_overrides_corrected(self):
        class A:
            def operation(self):
                pass

        class B(A):
            @overrides(A)
            def operation(self):
                return 42

        b = B()
        b.operation()

    def test_overrides_static_method(self):
        class A:
            @staticmethod
            def operation():
                pass

        class B(A):
            @staticmethod
            @overrides(A)
            def operation():
                return 42

        b = B()
        self.assertEqual(b.operation(), 42)
        self.assertEqual(B.operation(), 42)

    def test_deprecated_1(self):
        @deprecated
        def old_method(i: int) -> str:
            return str(i)

        with warnings.catch_warnings(record=True) as w:
            # Cause all warnings to always be triggered.
            warnings.simplefilter("always")
            # Trigger a warning.
            old_method(42)
            # Verify some things
            assert len(w) == 1
            assert issubclass(w[-1].category, DeprecationWarning)
            assert "deprecated" in str(w[-1].message)

    def test_deprecated_2(self):
        def old_method(i: int) -> str:
            return str(i)

        with warnings.catch_warnings(record=True) as w:
            # Cause all warnings to always be triggered.
            warnings.simplefilter("always")
            # Trigger a warning.
            old_method(42)
            # Verify some things
            assert not len(w) == 1

    def test_needs_refactoring_1(self):
        @needs_refactoring
        def old_method(i: int) -> str:
            return str(i)

        with warnings.catch_warnings(record=True) as w:
            # Cause all warnings to always be triggered.
            warnings.simplefilter("always")
            # Trigger a warning.
            old_method(42)
            # Verify some things
            assert len(w) == 1
            assert issubclass(w[-1].category, UserWarning)
            assert "refactoring" in str(w[-1].message)

    def test_needs_refactoring_2(self):
        def old_method(i: int) -> str:
            return str(i)

        with warnings.catch_warnings(record=True) as w:
            # Cause all warnings to always be triggered.
            warnings.simplefilter("always")
            # Trigger a warning.
            old_method(42)
            # Verify some things
            assert not len(w) == 1

    def test_dirty(self):
        @dirty
        def dirt(i: int) -> str:
            return str(i)

        with self.assertRaises(expected_exception=TooDirtyException):
            dirt(42)

    def test_unimplemented(self):
        @unimplemented
        def dirt(i: int) -> str:
            return str(i)

        with self.assertRaises(expected_exception=NotImplementedException):
            dirt(42)

    def test_timer(self):
        @timer
        def operation(i: int) -> str:
            return str(i)

        operation(42)

    def test_count_calls(self):
        @count_calls
        def operation(i: int) -> str:
            return str(i)

        operation(42)

    def test_validate_args(self):
        @validate_args(lambda x: (x > 42, f'Each arg should be > 42, but it was {x}.'))
        def some_calculation(a, b, c):
            return a + b + c

        some_calculation(43, 45, 50)
        with self.assertRaises(expected_exception=AssertionError):
            some_calculation(30, 40, 50)
        with self.assertRaises(expected_exception=AssertionError):
            some_calculation(c=30, a=40, b=50)

    def test_validate_args_instance_method(self):
        class MyClass:
            @validate_args(lambda x: (x > 0, f'Argument should be greater then 0, but it was {x}.'))
            def some_calculation(self, x: int) -> int:
                return x

        m = MyClass()
        m.some_calculation(1)
        m.some_calculation(42)
        with self.assertRaises(expected_exception=AssertionError):
            m.some_calculation(0)
        with self.assertRaises(expected_exception=AssertionError):
            m.some_calculation(-42)

    def test_require_not_none(self):
        @require_not_none
        def some_calculation(a, b, c):
            return a + b + c

        some_calculation(43, 0, -50)
        with self.assertRaises(expected_exception=AssertionError):
            some_calculation(30, None, 50)

    def test_require_not_empty_strings(self):
        @require_not_empty_strings
        def some_calculation(a, b, c):
            return a + b + c

        some_calculation('Hello ', 'W', 'orld   !')
        with self.assertRaises(expected_exception=AssertionError):
            some_calculation('Hello', '   ', 'World')
        with self.assertRaises(expected_exception=AssertionError):
            some_calculation('Hello', 4, 'World')
        with self.assertRaises(expected_exception=AssertionError):
            some_calculation('Hello', '4', None)



import unittest

# local file imports
from typing import Any

from pedantic import overrides, pedantic
from pedantic.class_decorators import pedantic_class, pedantic_class_require_docstring, trace_class, timer_class


class TestClassDecorators(unittest.TestCase):

    def test_pedantic_class_1(self):
        """Problem here: Argument 'a" in constructor doesn't have a type hint"""
        @pedantic_class
        class MyClass:
            def __init__(self, a) -> None:
                self.a = a

        with self.assertRaises(expected_exception=AssertionError):
            MyClass(a=42)

    def test_pedantic_class_2(self):
        """Problem here: Constructor must have type hint 'None'"""
        @pedantic_class
        class MyClass:
            def __init__(self, a: int):
                self.a = a

        with self.assertRaises(expected_exception=AssertionError):
            MyClass(a=42)

    def test_pedantic_class_3(self):
        """Problem here: Constructor must be called with kwargs"""
        @pedantic_class
        class MyClass:
            def __init__(self, a: int) -> None:
                self.a = a

        with self.assertRaises(expected_exception=AssertionError):
            MyClass(42)

    def test_pedantic_class_1_corrected(self):
        @pedantic_class
        class MyClass:
            def __init__(self, a: int) -> None:
                self.a = a

        MyClass(a=42)

    def test_multiple_methods_1(self):
        @pedantic_class
        class MyClass:
            def __init__(self, a: int) -> None:
                self.a = a

            def calc(self, b: int) -> int:
                return self.a - b

            def print(self, s: str) -> None:
                res = f'{self.a} and {s}'

        m = MyClass(a=5)
        m.calc(b=42)
        m.print(s='Hi')

    def test_multiple_methods_2(self):
        """Problem here: missing type hints"""
        @pedantic_class
        class MyClass:
            def __init__(self, a: int) -> None:
                self.a = a

            def calc(self, b: int) -> int:
                return self.a - b

            def print(self, s: str):
                print(f'{self.a} and {s}')

        m = MyClass(a=5)
        m.calc(b=42)
        with self.assertRaises(expected_exception=AssertionError):
            m.print(s='Hi')

    def test_multiple_methods_3(self):
        """Problem here: missing type hints"""
        @pedantic_class
        class MyClass:
            def __init__(self, a: int) -> None:
                self.a = a

            def calc(self, b) -> int:
                return self.a - b

            def print(self, s: str) -> None:
                res = f'{self.a} and {s}'

        m = MyClass(a=5)
        with self.assertRaises(expected_exception=AssertionError):
            m.calc(b=42)
        m.print(s='Hi')

    def test_multiple_methods_4(self):
        """Problem here: not called with kwargs"""
        @pedantic_class
        class MyClass:
            def __init__(self, a: int) -> None:
                self.a = a

            def calc(self, b: int) -> int:
                return self.a - b

            def print(self, s: str) -> None:
                res = f'{self.a} and {s}'

        m = MyClass(a=5)
        m.calc(b=42)
        with self.assertRaises(expected_exception=AssertionError):
            m.print('Hi')

    def test_generator_1(self):
        """Problem here: typo in type annotation string"""
        @pedantic_class
        class MyClass:
            def __init__(self, s: str) -> None:
                self.s = s

            def double(self, b: int) -> str:
                return self.s + str(b)

            @staticmethod
            def generator() -> 'MyClas':
                return MyClass(s='generated')

        with self.assertRaises(expected_exception=AssertionError):
            m = MyClass.generator()

    def test_generator_1_corrected(self):
        @pedantic_class
        class MyClass:
            def __init__(self, s: str) -> None:
                self.s = s

            def double(self, b: int) -> str:
                return self.s + str(b)

            @staticmethod
            def generator() -> 'MyClass':
                return MyClass(s='generated')

        m = MyClass.generator()
        m.double(b=42)

    def test_pedantic_class_require_docstring(self):
        @pedantic_class_require_docstring
        class MyClass:
            def __init__(self, s: str) -> None:
                """Constructor

                Args:
                    s (str): name
                """
                self.s = s

            def double(self, b: int) -> str:
                """some method

                Args:
                    b (int): magic number

                Returns:
                    str: cool stuff

                """
                return self.s + str(b)

            @staticmethod
            def generator() -> 'MyClass':
                """Static

                Returns:
                    MyClass: instance
                """
                return MyClass(s='generated')

        m = MyClass.generator()
        m.double(b=42)

    def test_pedantic_class_require_docstring_1(self):
        """Problem here: Typo in type annotation string"""
        @pedantic_class_require_docstring
        class MyClass:
            def __init__(self, s: str) -> None:
                """Constructor

                Args:
                    s (str): name
                """
                self.s = s

            def double(self, b: int) -> str:
                """some method

                Args:
                    b (int): magic number

                Returns:
                    str: cool stuff

                """
                return self.s + str(b)

            @staticmethod
            def generator() -> 'MyClas':
                """Static

                Returns:
                    MyClass: instance
                """
                return MyClass(s='generated')

        with self.assertRaises(expected_exception=AssertionError):
            m = MyClass.generator()

    def test_pedantic_class_require_docstring_2(self):
        """Problem here: Typo in docstring corresponding to type annotation string"""
        @pedantic_class_require_docstring
        class MyClass:
            def __init__(self, s: str) -> None:
                """Constructor

                Args:
                    s (str): name
                """
                self.s = s

            def double(self, b: int) -> str:
                """some method

                Args:
                    b (int): magic number

                Returns:
                    str: cool stuff

                """
                return self.s + str(b)

            @staticmethod
            def generator() -> 'MyClass':
                """Static

                Returns:
                    MyClas: instance
                """
                return MyClass(s='generated')

        with self.assertRaises(expected_exception=AssertionError):
            m = MyClass.generator()

    def test_pedantic_class_require_docstring_3(self):
        """Problem here: One docstring is wrong"""
        @pedantic_class_require_docstring
        class MyClass:
            def __init__(self, s: str) -> None:
                """Constructor

                Args:
                    s (str): name
                """
                self.s = s

            def double(self, b: int) -> str:
                """some method

                Args:
                    b (float): magic number

                Returns:
                    str: cool stuff

                """
                return self.s + str(b)

            @staticmethod
            def generator() -> 'MyClass':
                """Static

                Returns:
                    MyClass: instance
                """
                return MyClass(s='generated')

        m = MyClass.generator()
        with self.assertRaises(expected_exception=AssertionError):
            m.double(b=42)

    def test_trace_class(self):
        @trace_class
        class MyClass:
            def __init__(self, s: str) -> None:
                self.s = s

            def double(self, b: int) -> str:
                return self.s + str(b)

            @staticmethod
            def generator() -> 'MyClass':
                return MyClass(s='generated')

        m = MyClass.generator()
        m.double(b=42)

    def test_timer_class(self):
        @timer_class
        class MyClass:
            def __init__(self, s: str) -> None:
                self.s = s

            def double(self, b: int) -> str:
                return self.s + str(b)

            @staticmethod
            def generator() -> 'MyClass':
                return MyClass(s='generated')

        m = MyClass.generator()
        m.double(b=42)

    def test_pedantic_overloading_1(self):
        """Problem here: missing type hint for item"""
        @pedantic_class
        class MyClass(list):
            def __contains__(self, item) -> bool:
                return True

        m = MyClass()
        with self.assertRaises(expected_exception=AssertionError):
            print('something' in m)

    def test_pedantic_overloading_1_corrected(self):
        @pedantic_class
        class MyClass(list):
            def __contains__(self, item: str) -> bool:
                return True

        m = MyClass()
        print('something' in m)

    def test_pedantic_overloading_2(self):
        @pedantic_class
        class MyClass(list):
            def contains(self, item: str) -> bool:
                return True

        m = MyClass()
        print('something' in m)

    def test_type_annotation_string_1(self):
        """Problem here: typo in string type annotation"""
        @pedantic_class
        class MyClass:
            def compare(self, other: 'MyClas') -> bool:
                return False

        m = MyClass()
        with self.assertRaises(expected_exception=AssertionError):
            m.compare(other=m)

    def test_type_annotation_string_1_corrected(self):
        @pedantic_class
        class MyClass:
            def compare(self, other: 'MyClass') -> bool:
                return False

        m = MyClass()
        m.compare(other=m)

    def test_pedantic_class_docstring_1(self):
        """Problem here: syntax error in docstring"""

        with self.assertRaises(expected_exception=AssertionError):
            @pedantic_class
            class Foo:
                def __init__(self, a: int) -> None:
                    self.a = int

                def func(self, b: str) -> str:
                    """
                    Function with docstring syntax error below.
                    Args:
                        b (str):
                        simple string
                    Returns:
                        str: simple string
                    """
                    return b

                def bunk(self) -> int:
                    '''
                    Function with correct docstring.
                    Returns:
                        int: 42
                    '''
                    return 42

            foo = Foo(a=10)
            foo.func(b='bar')

    def test_pedantic_class_docstring_1_corrected(self):
        @pedantic_class
        class Foo:
            def __init__(self, a: int) -> None:
                self.a = int

            def func(self, b: str) -> str:
                """
                Function with docstring syntax error below.
                Args:
                    b (str): simple string
                Returns:
                    str: simple string
                """
                return b

            def bunk(self) -> int:
                '''
                Function with correct docstring.
                Returns:
                    int: 42
                '''
                return 42

        foo = Foo(a=10)
        foo.func(b='bar')

    def test_pedantic_class_overrides_1(self):
        @pedantic_class
        class Abstract:
            def func(self, b: str) -> str:
                pass

            def bunk(self) -> int:
                pass

        @pedantic_class
        class Foo(Abstract):
            def __init__(self, a: int) -> None:
                self.a = a

            @overrides(Abstract)
            def func(self, b: str) -> str:
                return b

            @overrides(Abstract)
            def bunk(self) -> int:
                return 42

        f = Foo(a=42)
        f.func(b='Hi')
        f.bunk()

    def test_pedantic_class_overrides_1_abstractmethod(self):
        from abc import ABC, abstractmethod

        @pedantic_class
        class Abstract(ABC):
            @abstractmethod
            def func(self, b: str) -> str:
                pass

            @abstractmethod
            def bunk(self) -> int:
                pass

        @pedantic_class
        class Foo(Abstract):
            def __init__(self, a: int) -> None:
                self.a = a

            @overrides(Abstract)
            def func(self, b: str) -> str:
                return b

            @overrides(Abstract)
            def bunk(self) -> int:
                return 42

        f = Foo(a=42)
        f.func(b='Hi')
        f.bunk()

    def test_pedantic_class_overrides_2(self):
        """Problem here: func != funcy"""
        @pedantic_class
        class Parent:
            def func(self, b: str) -> str:
                return b + b + b

            def bunk(self) -> int:
                return 42

        with self.assertRaises(expected_exception=AssertionError):
            @pedantic_class
            class Foo(Parent):
                def __init__(self, a: int) -> None:
                    self.a = a

                @overrides(Parent)
                def funcy(self, b: str) -> str:
                    return b

                @overrides(Parent)
                def bunk(self) -> int:
                    return self.a

            f = Foo(a=40002)
            f.func(b='Hi')
            f.bunk()

        p = Parent()
        p.func(b='Hi')
        p.bunk()

    def test_pedantic_class_overrides_3(self):
        """Problem here: type errors and call by args"""
        @pedantic_class
        class Abstract:
            def func(self, b: str) -> str:
                pass

            def bunk(self) -> int:
                pass

        @pedantic_class
        class Foo(Abstract):
            def __init__(self, a: int) -> None:
                self.a = a

            @overrides(Abstract)
            def func(self, b: str) -> str:
                return b

            @overrides(Abstract)
            def bunk(self) -> int:
                return self.a

        f = Foo(a=42)
        f.func(b='Hi')
        f.bunk()
        with self.assertRaises(expected_exception=AssertionError):
            f.func('Hi')
        with self.assertRaises(expected_exception=AssertionError):
            f2 = Foo(a=3.1415)
        f.a = 3.145
        with self.assertRaises(expected_exception=AssertionError):
            f.bunk()

    def test_pedantic_class_overrides_2_corrected(self):
        @pedantic_class
        class Parent:
            def func(self, b: str) -> str:
                return b + b + b

            def bunk(self) -> int:
                return 42

        @pedantic_class
        class Foo(Parent):
            def __init__(self, a: int) -> None:
                self.a = a

            @overrides(Parent)
            def func(self, b: str) -> str:
                return b

            @overrides(Parent)
            def bunk(self) -> int:
                return self.a

        f = Foo(a=40002)
        f.func(b='Hi')
        f.bunk()

        p = Parent()
        p.func(b='Hi')
        p.bunk()

    def test_static_method_1(self):
        """Problem here: Return type of "static_bar" should be Tuple[int, int]"""
        @pedantic_class
        class MyStaticClass:

            @staticmethod
            def double_func(a: int) -> int:
                x, y = MyStaticClass.static_bar()
                return x

            @staticmethod
            def static_bar() -> (int, int):  # this is wrong. Correct would be Tuple[int, int]
                return 0, 1

        with self.assertRaises(expected_exception=AssertionError):
            print(MyStaticClass.double_func(a=0))

    def test_property_1(self):
        @pedantic_class
        class MyClass(object):
            def __init__(self, some_arg: Any) -> None:
                self._some_attribute = some_arg

            @property
            def some_attribute(self) -> int:
                return self._some_attribute

            @some_attribute.setter
            def some_attribute(self, value: str) -> None:
                self._some_attribute = value

            def calc(self, value: float) -> float:
                return 2 * value

        with self.assertRaises(expected_exception=AssertionError):
            MyClass(42)

        m = MyClass(some_arg=42)
        with self.assertRaises(expected_exception=AssertionError):
            m.some_attribute = 100
        m.calc(value=42.0)

    def test_property_1(self):
        """Problem here: property setter and getter have incorrect type hints str != int != float"""
        @pedantic_class
        class MyClass(object):
            def __init__(self, some_arg: int) -> None:
                self._some_attribute = some_arg

            @property
            def some_attribute(self) -> float:
                return self._some_attribute

            @some_attribute.setter
            def some_attribute(self, value: str) -> None:
                self._some_attribute = value

            def calc(self, value: float) -> float:
                return 2 * value

        with self.assertRaises(expected_exception=AssertionError):
            MyClass(42)

        m = MyClass(some_arg=42)
        with self.assertRaises(expected_exception=AssertionError):
            m.some_attribute = 100

        with self.assertRaises(expected_exception=AssertionError):
            print(m.some_attribute)
        m.calc(value=42.0)
        with self.assertRaises(expected_exception=AssertionError):
            m.calc(value=42)

    def test_property_2(self):
        """Problems here: property setter and getter misses type hints"""

        @pedantic_class
        class MyClass(object):
            def __init__(self, some_arg: int) -> None:
                self._some_attribute = some_arg

            @property
            def some_attribute(self):
                return self._some_attribute

            @some_attribute.setter
            def some_attribute(self, value: int):
                self._some_attribute = value

            def calc(self, value: float) -> float:
                return 2 * value

        with self.assertRaises(expected_exception=AssertionError):
            MyClass(42)

        m = MyClass(some_arg=42)
        with self.assertRaises(expected_exception=AssertionError):
            m.some_attribute = 100

        with self.assertRaises(expected_exception=AssertionError):
            print(m.some_attribute)
        m.calc(value=42.0)
        with self.assertRaises(expected_exception=AssertionError):
            m.calc(value=42)

    def test_property_1_corrected(self):
        @pedantic_class
        class MyClass(object):
            def __init__(self, some_arg: int) -> None:
                self._some_attribute = some_arg

            @property
            def some_attribute(self) -> int:
                return self._some_attribute

            @some_attribute.setter
            def some_attribute(self, value: int) -> None:
                if value is None:
                    raise ValueError()
                self._some_attribute = value

            def calc(self, value: float) -> float:
                return 2 * value

        m = MyClass(some_arg=42)
        self.assertEqual(m.some_attribute, 42)
        m.some_attribute = 100
        self.assertEqual(m.some_attribute, 100)
        m.calc(value=42.0)


if __name__ == '__main__':
    # run single test
    test = TestClassDecorators()
    test.test_property_1()



import unittest
from typing import List

# local file imports
from pedantic.method_decorators import pedantic_require_docstring, pedantic


class TestRequireDocstringGoogleFormat(unittest.TestCase):

    def test_no_doc_string(self):
        """Problem here: No docstring"""
        @pedantic_require_docstring
        def calc(n: int, m: int, i: int) -> int:
            return n + m + i

        with self.assertRaises(expected_exception=AssertionError):
            calc(n=42, m=40, i=38)

    def test_one_line_doc_string(self):
        """Problem here: The docstring misses arguments and return value"""
        @pedantic_require_docstring
        def calc(n: int, m: int, i: int) -> int:
            """Returns the sum of the three args."""
            return n + m + i

        with self.assertRaises(expected_exception=AssertionError):
            calc(n=42, m=40, i=38)

    def test_one_line_doc_string_corrected(self):
        @pedantic_require_docstring
        def calc(n: int, m: int, i: int) -> int:
            """Returns the sum of the three args.

            Args:
                n (int): bla
                m (int): blub
                i (int): foo

            Returns:
                int: bar
            """
            return n + m + i

        calc(n=42, m=40, i=38)

    def test_google_docstring_1(self):
        """Problem here: List[str] != list"""
        @pedantic_require_docstring
        def calc(file_loc: str, print_cols: bool) -> List[str]:
            """Gets and prints the spreadsheet's header columns

            Args:
                file_loc (str): The file location of the spreadsheet
                print_cols (bool): A flag used to print the columns to the console
                    (default is False)

            Returns:
                list: a list of strings representing the header columns
            """

            return [file_loc, str(print_cols)]

        with self.assertRaises(expected_exception=AssertionError):
            calc(file_loc='Hi', print_cols=False)

    def test_google_docstring_2(self):
        @pedantic_require_docstring
        def calc(file_loc: str, print_cols: bool) -> List[str]:
            """Gets and prints the spreadsheet's header columns

            Args:
                file_loc (str): The file location of the spreadsheet
                print_cols (bool): A flag used to print the columns to the console
                    (default is False)

            Returns:
                List[str]: a list of strings representing the header columns
            """

            return [file_loc, str(print_cols)]

        calc(file_loc='Hi', print_cols=False)

    def test_google_docstring_3(self):
        @pedantic_require_docstring
        def calc(file_loc: str, print_cols: bool) -> List[str]:
            """Gets and prints the spreadsheet's header columns

            Args:
                file_loc (str): The file location of the spreadsheet
                print_cols (bool): A flag used to print the columns to the console
                    (default is False)

            Returns:
                List[str]: a list of strings representing the header columns
            """

            return [file_loc, str(print_cols)]

        calc(file_loc='Hi', print_cols=False)

    def test_google_docstring_4(self):
        """Problem here: Argument 'amount' is NOT a parameter of the function."""
        @pedantic_require_docstring
        def calc(file_loc: str, print_cols: bool) -> List[str]:
            """Gets and prints the spreadsheet's header columns

            Args:
                file_loc (str): The file location of the spreadsheet
                print_cols (bool): A flag used to print the columns to the console
                    (default is False)
                amount (int): THIS ARGUMENT IS NOT TAKEN BY THE FUNCTION

            Returns:
                List[str]: a list of strings representing the header columns
            """

            return [file_loc, str(print_cols)]

        with self.assertRaises(expected_exception=AssertionError):
            calc(file_loc='Hi', print_cols=False)

    def test_google_docstring_corrected(self):
        @pedantic_require_docstring
        def calc(file_loc: str, print_cols: bool, amount: int) -> List[str]:
            """Gets and prints the spreadsheet's header columns

            Args:
                file_loc (str): The file location of the spreadsheet
                print_cols (bool): A flag used to print the columns to the console
                    (default is False)
                amount (int): now it is

            Returns:
                List[str]: a list of strings representing the header columns
            """

            return [file_loc, str(print_cols), str(amount)]

        calc(file_loc='Hi', print_cols=False, amount=42)

    def test_google_no_args_keyword(self):
        """Problem here: Docstring misses the keyword 'Args:' before the arguments listed."""
        @pedantic_require_docstring
        def calc(file_loc: str, print_cols: bool) -> list:
            """Gets and prints the spreadsheet's header columns

            file_loc (str): The file location of the spreadsheet
            print_cols (bool): A flag used to print the columns to the console
                (default is False)

            Returns:
                list: a list of strings representing the header columns
            """

            return [file_loc, str(print_cols)]

        with self.assertRaises(expected_exception=AssertionError):
            calc(file_loc='Hi', print_cols=False)

    def test_google_no_return_keyword(self):
        """Problem here: Docstring misses the keyword 'Returns:' before the return value."""
        @pedantic_require_docstring
        def calc(file_loc: str, print_cols: bool) -> list:
            """Gets and prints the spreadsheet's header columns

            Args:
                file_loc (str): The file location of the spreadsheet
                print_cols (bool): A flag used to print the columns to the console
                    (default is False)

            list: a list of strings representing the header columns
            """

            return [file_loc, str(print_cols)]

        with self.assertRaises(expected_exception=AssertionError):
            calc(file_loc='Hi', print_cols=False)

    def test_keep_it_simple(self):
        @pedantic_require_docstring
        def calc() -> None:
            """Gets and prints the spreadsheet's header columns"""
            a = 'hi'

        calc()

    def test_keep_it_simple_2(self):
        """Problem here: docstring misses argument"""
        @pedantic_require_docstring
        def calc(name: str) -> None:
            """Gets and prints the spreadsheet's header columns"""
            a = 'hi ' + name

        with self.assertRaises(expected_exception=AssertionError):
            calc(name='maria')

    def test_keep_it_simple_2_corrected(self):
        @pedantic_require_docstring
        def calc(name: str) -> None:
            """Gets and prints the spreadsheet's header columns

            Args:
                name (str): the name
            """
            a = 'hi ' + name

        calc(name='maria')

    def test_undocumented_arg(self):
        """Problem here: docstring misses argument"""
        @pedantic_require_docstring
        def calc(file_loc: str, print_cols: bool, number: int) -> List[str]:
            """Gets and prints the spreadsheet's header columns

            Args:
                file_loc (str): The file location of the spreadsheet
                print_cols (bool): A flag used to print the columns to the console
                    (default is False)

            Returns:
                List[str]: a list of strings representing the header columns
            """

            return [file_loc, str(print_cols), str(number)]

        with self.assertRaises(expected_exception=AssertionError):
            calc(file_loc='Hi', print_cols=False, number=42)

    def test_undocumented_arg_corrected(self):
        @pedantic_require_docstring
        def calc(file_loc: str, print_cols: bool, number: int) -> List[str]:
            """Gets and prints the spreadsheet's header columns

            Args:
                file_loc (str): The file location of the spreadsheet
                print_cols (bool): A flag used to print the columns to the console
                    (default is False)
                number (int): magic number

            Returns:
                List[str]: a list of strings representing the header columns
            """

            return [file_loc, str(print_cols), str(number)]

        calc(file_loc='Hi', print_cols=False, number=42)

    def test_restructured_text_1(self):
        """Problem here: reStructured Text docstring format cannot be parsed"""
        @pedantic_require_docstring
        def calc(file_loc: str, print_cols: bool) -> List[str]:
            """Gets and prints the spreadsheet's header columns

            :param file_loc: The file location of the spreadsheet
            :type file_loc: str
            :param print_cols: A flag used to print the columns to the console
            :type print_cols: bool
            :returns: a list of strings representing the header column
            :rtype: List[str]
            """
            return [file_loc, str(print_cols)]

        with self.assertRaises(expected_exception=AssertionError):
            calc(file_loc='Hi', print_cols=False)

    def test_return_value_1(self):
        """Problem here: There is a return value documented, but nothing is returned."""
        @pedantic_require_docstring
        def calc(file_loc: str, print_cols: bool):
            """Gets and prints the spreadsheet's header columns

            Args:
                file_loc (str): The file location of the spreadsheet
                print_cols (bool): A flag used to print the columns to the console
                    (default is False)

            Returns:
                list: a list of strings representing the header columns
            """

            a = [file_loc, str(print_cols)]

        with self.assertRaises(expected_exception=AssertionError):
            calc(file_loc='Hi', print_cols=False)

    def test_return_value_11(self):
        """Problem here: There is a return value documented, but nothing is returned."""
        @pedantic_require_docstring
        def calc(file_loc: str, print_cols: bool) -> None:
            """Gets and prints the spreadsheet's header columns

            Args:
                file_loc (str): The file location of the spreadsheet
                print_cols (bool): A flag used to print the columns to the console
                    (default is False)

            Returns:
                list: a list of strings representing the header columns
            """

            a = [file_loc, str(print_cols)]

        with self.assertRaises(expected_exception=AssertionError):
            calc(file_loc='Hi', print_cols=False)

    def test_return_value_1_corrected(self):
        @pedantic_require_docstring
        def calc(file_loc: str, print_cols: bool) -> None:
            """Gets and prints the spreadsheet's header columns

            Args:
                file_loc (str): The file location of the spreadsheet
                print_cols (bool): A flag used to print the columns to the console
                    (default is False)
            """

            a = [file_loc, str(print_cols)]
        calc(file_loc='Hi', print_cols=False)

    def test_return_value_2(self):
        """Problem here: the return value is not documented"""
        @pedantic_require_docstring
        def calc(file_loc: str, print_cols: bool) -> List[str]:
            """Gets and prints the spreadsheet's header columns

            Args:
                file_loc (str): The file location of the spreadsheet
                print_cols (bool): A flag used to print the columns to the console (default is False)

            Returns:
            """

            return [file_loc, str(print_cols)]

        with self.assertRaises(expected_exception=AssertionError):
            # the error message here is actually wrong due to the behavior of the docstring-parser package
            calc(file_loc='Hi', print_cols=False)

    def test_return_value_2_corrected(self):
        @pedantic_require_docstring
        def calc(file_loc: str, print_cols: bool) -> List[str]:
            """Gets and prints the spreadsheet's header columns

            Args:
                file_loc (str): The file location of the spreadsheet
                print_cols (bool): A flag used to print the columns to the console
                    (default is False)

            Returns:
                List[str]: results
            """

            return [file_loc, str(print_cols)]

        calc(file_loc='Hi', print_cols=False)

    def test_return_value_3(self):
        """Problem here: the return value is not documented"""
        @pedantic_require_docstring
        def calc(file_loc: str, print_cols: bool) -> List[str]:
            """Gets and prints the spreadsheet's header columns

            Args:
                file_loc (str): The file location of the spreadsheet
                print_cols (bool): A flag used to print the columns to the console
                    (default is False)
            """

            return [file_loc, str(print_cols)]

        with self.assertRaises(expected_exception=AssertionError):
            calc(file_loc='Hi', print_cols=False)

    def test_wrong_format_1(self):
        """Problem here: Currently, only Google docstrings are supported"""
        class MyText:
            text = 'hi'

            @pedantic_require_docstring
            def __contains__(self, substring: str) -> bool:
                """
                Checks if contains substring.
                Overriding __contains__ build in functions allows to use the 'in' operator blah readability

                Example:
                my_text = MyText('abc')
                if 'ab' in my_text -> true
                :param: substring: substring
                :return: True if substring is stored, False otherwise.
                """
                return substring in self.text

        my_text = MyText()
        with self.assertRaises(expected_exception=AssertionError):
            assert 'ab' not in my_text
            assert 'i' in my_text

    def test_pedantic_1(self):
        with self.assertRaises(expected_exception=AssertionError):
            @pedantic
            def calc(a: int, b: float, c: str) -> str:
                """Returns some cool string

                Args:
                    a (int): dfnoghbn
                    b (float): fubsuibgi

                Returns:
                    str: fshgimhj
                """
                return str(a) + str(b) + c

            calc(a=42, b=3.14, c='hi')

    def test_pedantic_1_corrected(self):
        @pedantic
        def calc(a: int, b: float, c: str) -> str:
            """Returns some cool string

            Args:
                a (int): dfnoghbn
                b (float): fubsuibgi
                c (str): dfgnhoif

            Returns:
                str: fshgimhj
            """
            return str(a) + str(b) + c

        calc(a=42, b=3.14, c='hi')


if __name__ == '__main__':
    t = TestRequireDocstringGoogleFormat()
    t.test_return_value_2()



import unittest
from typing import List, Tuple, Callable, Any, Optional, Union, Dict, Set, FrozenSet

# local file imports
from pedantic.method_decorators import pedantic


class TestDecoratorRequireKwargsAndTypeCheck(unittest.TestCase):

    def test_no_kwargs_1(self):
        """Problem here: function is not called with keyword arguments"""
        @pedantic
        def calc(n: int, m: int, i: int) -> int:
            return n + m + i

        with self.assertRaises(expected_exception=AssertionError):
            calc(42, 40, 38)

    def test_no_kwargs_2(self):
        """Problem here: function is not called with keyword arguments"""
        @pedantic
        def calc(n: int, m: int, i: int) -> int:
            return n + m + i

        with self.assertRaises(expected_exception=AssertionError):
            calc(42, m=40, i=38)

    def test_no_kwargs_1_2_corrected(self):
        @pedantic
        def calc(n: int, m: int, i: int) -> int:
            return n + m + i

        calc(n=42, m=40, i=38)

    def test_nested_type_hints_1(self):
        """Problem here: actual return type doesn't match return type annotation"""
        @pedantic
        def calc(n: int) -> List[List[float]]:
            return [0.0 * n]

        with self.assertRaises(expected_exception=AssertionError):
            calc(n=42)

    def test_nested_type_hints_1_corrected(self):
        @pedantic
        def calc(n: int) -> List[List[float]]:
            return [[0.0 * n]]

        calc(n=42)

    def test_nested_type_hints_2(self):
        """Problem here: int != float"""
        @pedantic
        def calc(n: int) -> List[Tuple[float, str]]:
            return [(n, str(n))]

        with self.assertRaises(expected_exception=AssertionError):
            calc(n=42)

    def test_nested_type_hints_2_corrected(self):
        @pedantic
        def calc(n: int) -> List[Tuple[int, str]]:
            return [(n, str(n))]

        @pedantic
        def calc_2(n: float) -> List[Tuple[float, str]]:
            return [(n, str(n))]

        calc(n=42)
        calc_2(n=42.0)

    def test_nested_type_hints_3(self):
        """Problem here: inner function actually returns Tuple[int, str]"""
        @pedantic
        def calc(n: int) -> Callable[[int, float], Tuple[float, str]]:
            @pedantic
            def f(x: int, y: float) -> Tuple[float, str]:
                return n * x, str(y)
            return f

        with self.assertRaises(expected_exception=AssertionError):
            calc(n=42)(x=3, y=3.14)

    def test_nested_type_hints_3_corrected(self):
        @pedantic
        def calc(n: int) -> Callable[[int, float], Tuple[int, str]]:
            @pedantic
            def f(x: int, y: float) -> Tuple[int, str]:
                return n * x, str(y)

            return f

        calc(n=42)(x=3, y=3.14)

    def test_nested_type_hints_4(self):
        """Problem here: return type is actually float"""
        @pedantic
        def calc(n: List[List[float]]) -> int:
            return n[0][0]

        with self.assertRaises(expected_exception=AssertionError):
            calc(n=[[42.0]])

    def test_nested_type_hints_corrected(self):
        @pedantic
        def calc(n: List[List[float]]) -> int:
            return int(n[0][0])

        calc(n=[[42.0]])

    def test_nested_type_hints_5(self):
        """Problem here: Tuple[float, str] != Tuple[float, float]"""

        @pedantic
        def calc(n: int) -> Callable[[int, float], Tuple[float, str]]:
            @pedantic
            def f(x: int, y: float) -> Tuple[float, float]:
                return n * float(x), y
            return f

        with self.assertRaises(expected_exception=AssertionError):
            calc(n=42)

    def test_nested_type_hints_5_corrected(self):
        @pedantic
        def calc(n: int) -> Callable[[int, float], Tuple[float, float]]:
            @pedantic
            def f(x: int, y: float) -> Tuple[float, float]:
                return n * float(x), y
            return f

        calc(n=42)

    def test_missing_type_hint_1(self):
        """Problem here: type hint for n missed"""
        @pedantic
        def calc(n) -> float:
            return 42.0 * n

        with self.assertRaises(expected_exception=AssertionError):
            calc(n=42)

    def test_missing_type_hint_1_corrected(self):
        @pedantic
        def calc(n: int) -> float:
            return 42.0 * n

        calc(n=42)

    def test_missing_type_hint_2(self):
        """Problem here: Return type annotation missed"""
        @pedantic
        def calc(n: int):
            return 'Hi' + str(n)

        with self.assertRaises(expected_exception=AssertionError):
            calc(n=42)

    def test_missing_type_hint_2_corrected(self):
        @pedantic
        def calc(n: int) -> str:
            return 'Hi' + str(n)

        calc(n=42)

    def test_missing_type_hint_3(self):
        """Problem here: type hint for i missed"""
        @pedantic
        def calc(n: int, m: int, i) -> int:
            return n + m + i

        with self.assertRaises(expected_exception=AssertionError):
            calc(n=42, m=40, i=38)

    def test_missing_type_hint_3_corrected(self):
        @pedantic
        def calc(n: int, m: int, i: int) -> int:
            return n + m + i

        calc(n=42, m=40, i=38)

    def test_all_ok_2(self):
        @pedantic
        def calc(n: int, m: int, i: int) -> str:
            return str(n + m + i)

        calc(n=42, m=40, i=38)

    def test_all_ok_3(self):
        @pedantic
        def calc(n: int, m: int, i: int) -> None:
            str(n + m + i)

        calc(n=42, m=40, i=38)

    def test_all_ok_4(self):
        @pedantic
        def calc(n: int) -> List[List[int]]:
            return [[n]]

        calc(n=42)

    def test_all_ok_5(self):
        @pedantic
        def calc(n: int) -> List[Tuple[float, str]]:
            return [(float(n), str(n))]

        calc(n=42)

    def test_all_ok_6(self):
        @pedantic
        def calc(n: int) -> Callable[[int, float], Tuple[float, str]]:
            @pedantic
            def f(x: int, y: float) -> Tuple[float, str]:
                return n * float(x), str(y)
            return f

        calc(n=42)(x=72, y=3.14)

    def test_all_ok_7(self):
        @pedantic
        def calc(n: List[List[float]]) -> Any:
            return n[0][0]

        calc(n=[[42.0]])

    def test_all_ok_8(self):
        @pedantic
        def calc(n: int) -> Callable[[int, float], Tuple[float, str]]:
            @pedantic
            def f(x: int, y: float) -> Tuple[float, str]:
                return n * float(x), str(y)

            return f

        calc(n=42)(x=3, y=3.14)

    def test_wrong_type_hint_1(self):
        """Problem here: str != int"""
        @pedantic
        def calc(n: int, m: int, i: int) -> str:
            return n + m + i

        with self.assertRaises(expected_exception=AssertionError):
            calc(n=42, m=40, i=38)

    def test_wrong_type_hint_1_corrected(self):
        @pedantic
        def calc(n: int, m: int, i: int) -> str:
            return str(n + m + i)

        calc(n=42, m=40, i=38)

    def test_wrong_type_hint_2(self):
        """Problem here: str != int"""
        @pedantic
        def calc(n: int, m: int, i: str) -> int:
            return n + m + i

        with self.assertRaises(expected_exception=AssertionError):
            calc(n=42, m=40, i=38)

    def test_wrong_type_hint_2_corrected(self):
        @pedantic
        def calc(n: int, m: int, i: str) -> int:
            return n + m + int(i)

        calc(n=42, m=40, i='38')

    def test_wrong_type_hint_3(self):
        """Problem here: None != int"""
        @pedantic
        def calc(n: int, m: int, i: int) -> None:
            return n + m + i

        with self.assertRaises(expected_exception=AssertionError):
            calc(n=42, m=40, i=38)

    def test_wrong_type_hint_corrected(self):
        @pedantic
        def calc(n: int, m: int, i: int) -> None:
            n + m + i

        calc(n=42, m=40, i=38)

    def test_wrong_type_hint_4(self):
        """Problem here: None != int"""
        @pedantic
        def calc(n: int, m: int, i: int) -> int:
            n + m + i

        with self.assertRaises(expected_exception=AssertionError):
            calc(n=42, m=40, i=38)

    def test_wrong_type_hint_4_corrected(self):
        @pedantic
        def calc(n: int, m: int, i: int) -> int:
            return n + m + i

        calc(n=42, m=40, i=38)

    def test_none_1(self):
        """Problem here: None is not accepted"""
        @pedantic
        def calc(n: int, m: int, i: int) -> int:
            return n + m + i

        with self.assertRaises(expected_exception=AssertionError):
            calc(n=42, m=40, i=None)

    def test_none_2(self):
        @pedantic
        def calc(n: int, m: int, i: Optional[int]) -> int:
            return n + m + i if i is not None else n + m

        calc(n=42, m=40, i=None)

    def test_none_3(self):
        @pedantic
        def calc(n: int, m: int, i: Union[int, None]) -> int:
            return n + m + i if i is not None else n + m

        calc(n=42, m=40, i=None)

    def test_none_4(self):
        """Problem here: function may return None"""
        @pedantic
        def calc(n: int, m: int, i: Union[int, None]) -> int:
            return n + m + i if i is not None else None

        with self.assertRaises(expected_exception=AssertionError):
            calc(n=42, m=40, i=None)

    def test_none_5(self):
        @pedantic
        def calc(n: int, m: int, i: Union[int, None]) -> Optional[int]:
            return n + m + i if i is not None else None

        calc(n=42, m=40, i=None)

    def test_inheritance_1(self):
        class A:
            pass

        class B(A):
            pass

        @pedantic
        def calc(a: A) -> str:
            return str(a)

        calc(a=A())
        calc(a=B())

    def test_inheritance_2(self):
        """Problem here: A is not a subtype of B"""
        class A:
            pass

        class B(A):
            pass

        @pedantic
        def calc(a: B) -> str:
            return str(a)

        calc(a=B())
        with self.assertRaises(expected_exception=AssertionError):
            calc(a=A())

    def test_instance_method_1(self):
        class A:
            @pedantic
            def calc(self, i: int) -> str:
                return str(i)

        a = A()
        a.calc(i=42)

    def test_instance_method_2(self):
        """Problem here: 'i' has no type annotation"""
        class A:
            @pedantic
            def calc(self, i) -> str:
                return str(i)

        a = A()
        with self.assertRaises(expected_exception=AssertionError):
            a.calc(i=42)

    def test_instance_method_2_corrected(self):
        class A:
            @pedantic
            def calc(self, i: int) -> str:
                return str(i)

        a = A()
        a.calc(i=42)

    def test_instance_method_3(self):
        """Problem here: float != int"""
        class A:
            @pedantic
            def calc(self, i: float) -> str:
                return str(i)

        a = A()
        with self.assertRaises(expected_exception=AssertionError):
            a.calc(i=42)

    def test_instance_method_3_corrected(self):
        class A:
            @pedantic
            def calc(self, i: float) -> str:
                return str(i)

        a = A()
        a.calc(i=42.0)

    def test_instance_method_4(self):
        """Problem here: instance methods is not called with kwargs"""
        class A:
            @pedantic
            def calc(self, i: int) -> str:
                return str(i)

        a = A()
        with self.assertRaises(expected_exception=AssertionError):
            a.calc(42)

    def test_instance_method_5(self):
        """Problem here: instance methods is not called with kwargs"""
        class A:
            @pedantic
            def calc(self, i: int) -> str:
                return str(i)

        a = A()
        a.calc(i=42)

    def test_lambda_1(self):
        """Lambda expressions cannot be typed hinted. So this leads to an error."""
        @pedantic
        def calc(i: float) -> Callable[[float], str]:
            return lambda x: str(x * i)

        with self.assertRaises(expected_exception=AssertionError):
            calc(i=42.0)(10.0)

    def test_lambda_2(self):
        """Even this is not expected by the type checker. Only test_lambda_3 has the 'correct' syntax"""
        @pedantic
        def calc(i: float) -> Callable[[float], str]:
            res: Callable[[float], str] = lambda x: str(x * i)
            return res

        with self.assertRaises(expected_exception=AssertionError):
            calc(i=42.0)(10.0)

    def test_lambda_3(self):
        @pedantic
        def calc(i: float) -> Callable[[float], str]:
            def res(x: float) -> str:
                return str(x * i)
            return res

        calc(i=42.0)(10.0)

    def test_lambda_4(self):
        """Problem here: inner function: int != float"""
        @pedantic
        def calc(i: float) -> Callable[[float], str]:
            def res(x: int) -> str:
                return str(x * i)
            return res

        with self.assertRaises(expected_exception=AssertionError):
            calc(i=42.0)(x=10)

    def test_lambda_4_almost_corrected(self):
        """Problem here: float != str"""
        @pedantic
        def calc(i: float) -> Callable[[float], str]:
            @pedantic
            def res(x: int) -> str:
                return str(x * i)
            return res

        with self.assertRaises(expected_exception=AssertionError):
            calc(i=42.0)(x=10)

    def test_lambda_4_almost_corrected_2(self):
        @pedantic
        def calc(i: float) -> Callable[[int], str]:
            @pedantic
            def res(x: int) -> str:
                return str(x * i)
            return res

        calc(i=42.0)(x=10)

    def test_lambda_5(self):
        """Problem here: float != int"""
        @pedantic
        def calc(i: float) -> Callable[[float], str]:
            @pedantic
            def res(x: float) -> str:
                return str(x * i)
            return res

        with self.assertRaises(expected_exception=AssertionError):
            calc(i=42.0)(x=10)

    def test_lambda_corrected(self):
        @pedantic
        def calc(i: float) -> Callable[[float], str]:
            @pedantic
            def res(x: float) -> str:
                return str(x * i)

            return res

        calc(i=42.0)(x=10.0)

    def test_tuple_without_args(self):
        """Problem here: Tuple has no type arguments"""
        @pedantic
        def calc(i: Tuple) -> str:
            return str(i)

        with self.assertRaises(expected_exception=AssertionError):
            calc(i=(42.0, 43, 'hi'))

    def test_tuple_without_args_corrected(self):
        @pedantic
        def calc(i: Tuple[Any, ...]) -> str:
            return str(i)

        calc(i=(42.0, 43, 'hi'))

    def test_callable_without_args(self):
        """Problem here: Callable has no type arguments"""
        @pedantic
        def calc(i: Callable) -> str:
            return str(i(' you'))

        with self.assertRaises(expected_exception=AssertionError):
            calc(i=lambda x: (42.0, 43, 'hi', x))

    def test_callable_without_args_almost_corrected(self):
        """Problem here: lambda expressions cannot be type hinted. So don't use it"""
        @pedantic
        def calc(i: Callable[[Any], Tuple[Any, ...]]) -> str:
            return str(i(x=' you'))

        with self.assertRaises(expected_exception=AssertionError):
            calc(i=lambda x: (42.0, 43, 'hi', x))

    def test_callable_without_args_corrected(self):
        @pedantic
        def calc(i: Callable[[Any], Tuple[Any, ...]]) -> str:
            return str(i(x=' you'))

        @pedantic
        def arg(x: Any) -> Tuple[Any, ...]:
            return 42.0, 43, 'hi', x
        calc(i=arg)

    def test_list_without_args(self):
        """Problem here: List has no type arguments"""
        @pedantic
        def calc(i: List) -> Any:
            return [i]

        with self.assertRaises(expected_exception=AssertionError):
            calc(i=[42.0, 43, 'hi'])

    def test_list_without_args_corrected(self):
        @pedantic
        def calc(i: List[Any]) -> List[List[Any]]:
            return [i]

        calc(i=[42.0, 43, 'hi'])

    def test_ellipsis_in_callable_1(self):
        @pedantic
        def calc(i: Callable[..., int]) -> int:
            return i()

        @pedantic
        def call() -> int:
            return 42

        calc(i=call)

    def test_ellipsis_in_callable_2(self):
        @pedantic
        def calc(i: Callable[..., int]) -> int:
            return i(x=3.14, y=5)

        @pedantic
        def call(x: float, y: int) -> int:
            return 42

        calc(i=call)

    def test_ellipsis_in_callable_3(self):
        """Problem here: call to "call" misses one argument"""
        @pedantic
        def calc(i: Callable[..., int]) -> int:
            return i(x=3.14)

        @pedantic
        def call(x: float, y: int) -> int:
            return 42

        with self.assertRaises(expected_exception=AssertionError):
            calc(i=call)

    def test_optional_args_1(self):
        @pedantic
        def calc(a: int, b: int = 42) -> int:
            return a + b

        calc(a=2)

    def test_optional_args_2(self):
        @pedantic
        def calc(a: int = 3, b: int = 42, c: float = 5.0) -> float:
            return a + b + c

        calc()
        calc(a=1)
        calc(b=1)
        calc(c=1.0)
        calc(a=1, b=1)
        calc(a=1, c=1.0)
        calc(b=1, c=1.0)
        calc(a=1, b=1, c=1.0)

    def test_optional_args_3(self):
        """Problem here: optional argument c: 5 is not a float"""
        @pedantic
        def calc(a: int = 3, b: int = 42, c: float = 5) -> float:
            return a + b + c

        with self.assertRaises(expected_exception=AssertionError):
            calc()

    def test_optional_args_3_corrected(self):
        @pedantic
        def calc(a: int = 3, b: int = 42, c: float = 5.0) -> float:
            return a + b + c

        calc()

    def test_optional_args_4(self):
        class MyClass:
            @pedantic
            def foo(self, a: int, b: Optional[int] = 1) -> int:
                return a + b

        myclass = MyClass()
        myclass.foo(a=10)

    def test_optional_args_5(self):
        @pedantic
        def calc(d: Optional[Dict[int, int]] = None) -> Optional[int]:
            if d is None:
                return None
            return sum(d.keys())

        calc(d=None)
        calc()
        calc(d={42: 3})

        with self.assertRaises(expected_exception=AssertionError):
            calc(d={42: 3.14})

    def test_optional_args_6(self):
        """"Problem here: str != int"""
        @pedantic
        def calc(d: int = 42) -> int:
            return int(d)

        calc(d=99999)
        with self.assertRaises(expected_exception=AssertionError):
            calc(d='999999')

    def test_enum_1(self):
        """Problem here: Type hint for a should be MyEnum instead of MyEnum.SEQUENCEFLOW"""
        from enum import Enum

        class MyEnum(Enum):
            STARTEVENT = 'startEvent'
            ACTIVITY = 'task'
            SEQUENCEFLOW = 'sequenceFlow'

        class MyClass:
            @pedantic
            def operation(self, a: MyEnum.SEQUENCEFLOW) -> None:
                print(a)

        m = MyClass()
        with self.assertRaises(expected_exception=AssertionError):
            m.operation(a=MyEnum.SEQUENCEFLOW)

    def test_enum_1_corrected(self):
        from enum import Enum

        class MyEnum(Enum):
            STARTEVENT = 'startEvent'
            ACTIVITY = 'task'
            SEQUENCEFLOW = 'sequenceFlow'
            ENDEVENT = 'endEvent'
            ID = 'id'
            NAME = 'name'

        @pedantic
        def operation(a: MyEnum) -> None:
            print(a)

        operation(a=MyEnum.SEQUENCEFLOW)

    def test_sloppy_types_dict(self):
        """Problem here: use typing.Dict instead of dict"""
        @pedantic
        def operation(d: dict) -> int:
            return len(d.keys())

        with self.assertRaises(expected_exception=AssertionError):
            operation(d={1: 1, 2: 2})

    def test_sloppy_types_dict_almost_corrected(self):
        """Problem here: typing.Dict misses type arguments"""
        @pedantic
        def operation(d: Dict) -> int:
            return len(d.keys())

        with self.assertRaises(expected_exception=AssertionError):
            operation(d={1: 1, 2: 2})

    def test_sloppy_types_dict_corrected(self):
        @pedantic
        def operation(d: Dict[int, int]) -> int:
            return len(d.keys())

        operation(d={1: 1, 2: 2})

    def test_sloppy_types_list(self):
        """Problem here: use typing.List instead of list"""
        @pedantic
        def operation(d: list) -> int:
            return len(d)

        with self.assertRaises(expected_exception=AssertionError):
            operation(d=[1, 2, 3, 4])

    def test_sloppy_types_list_almost_corrected(self):
        """Problem here: typing.List misses type argument"""
        @pedantic
        def operation(d: List) -> int:
            return len(d)

        with self.assertRaises(expected_exception=AssertionError):
            operation(d=[1, 2, 3, 4])

    def test_sloppy_types_list_corrected(self):
        @pedantic
        def operation(d: List[int]) -> int:
            return len(d)

        operation(d=[1, 2, 3, 4])

    def test_sloppy_types_tuple(self):
        """Problem here: use typing.Tuple instead of tuple"""
        @pedantic
        def operation(d: tuple) -> int:
            return len(d)

        with self.assertRaises(expected_exception=AssertionError):
            operation(d=(1, 2, 3))

    def test_sloppy_types_tuple_almost_corrected(self):
        """Problem here: typing.Tuple misses type arguments"""
        @pedantic
        def operation(d: Tuple) -> int:
            return len(d)

        with self.assertRaises(expected_exception=AssertionError):
            operation(d=(1, 2, 3))

    def test_sloppy_types_tuple_corrected(self):
        @pedantic
        def operation(d: Tuple[int, int, int]) -> int:
            return len(d)

        operation(d=(1, 2, 3))

    def test_sloppy_types_set(self):
        """Problem here: use typing.Set instead of set"""
        @pedantic
        def operation(d: set) -> int:
            return len(d)

        with self.assertRaises(expected_exception=AssertionError):
            operation(d={1, 2, 3})

    def test_sloppy_types_set_almost_corrected(self):
        """Problem here: typing.Set misses type argument"""
        @pedantic
        def operation(d: Set) -> int:
            return len(d)

        with self.assertRaises(expected_exception=AssertionError):
            operation(d={1, 2, 3})

    def test_sloppy_types_set_corrected(self):
        @pedantic
        def operation(d: Set[int]) -> int:
            return len(d)

        operation(d={1, 2, 3})

    def test_sloppy_types_frozenset(self):
        """Problem here: use typing.FrozenSet instead of frozenset"""
        @pedantic
        def operation(d: frozenset) -> int:
            return len(d)

        with self.assertRaises(expected_exception=AssertionError):
            operation(d=frozenset({1, 2, 3}))

    def test_sloppy_types_frozenset_almost_corrected(self):
        """Problem here: typing.FrozenSet misses type argument"""
        @pedantic
        def operation(d: FrozenSet) -> int:
            return len(d)

        with self.assertRaises(expected_exception=AssertionError):
            operation(d=frozenset({1, 2, 3}))

    def test_sloppy_types_frozenset_corrected(self):
        @pedantic
        def operation(d: FrozenSet[int]) -> int:
            return len(d)

        operation(d=frozenset({1, 2, 3}))

    def test_type_list(self):
        """Problem here: tuple != list"""
        @pedantic
        def calc(ls: List[Any]) -> int:
            return len(ls)

        with self.assertRaises(expected_exception=AssertionError):
            calc(ls=(1, 2, 3))

    def test_type_list_corrected(self):
        @pedantic
        def calc(ls: Tuple[Any, ...]) -> int:
            return len(ls)

        calc(ls=(1, 2, 3))

    def test_any(self):
        @pedantic
        def calc(ls: List[Any]) -> Dict[int, Any]:
            return {i: ls[i] for i in range(0, len(ls))}

        calc(ls=[1, 2, 3])
        calc(ls=[1.11, 2.0, 3.0])
        calc(ls=['1', '2', '3'])
        calc(ls=[10.5, '2', (3, 4, 5)])


if __name__ == '__main__':
    # run a specific unit test
    test = TestDecoratorRequireKwargsAndTypeCheck()
    test.test_instance_method_5()



"""From: https://stackoverflow.com/a/55504010/10975692"""

import inspect
import typing
import collections

typing_protocol = typing.Protocol if hasattr(typing, 'Protocol') else typing._Protocol


def is_instance(obj: typing.Any, type_: typing.Any) -> bool:
    """The main function of this module."""

    if type_.__module__ == 'typing':
        if is_qualified_generic(type_):
            base_generic = get_base_generic(type_)
        else:
            base_generic = type_
        name = _get_name(base_generic)

        try:
            validator = _SPECIAL_INSTANCE_CHECKERS[name]
        except KeyError:
            pass
        else:
            return validator(obj, type_)

    if is_base_generic(type_):
        python_type = _get_python_type(type_)
        return isinstance(obj, python_type)

    if is_qualified_generic(type_):
        python_type = _get_python_type(type_)
        if not isinstance(obj, python_type):
            return False

        base = get_base_generic(type_)
        validator = _ORIGIN_TYPE_CHECKERS[base]

        type_args = get_subtypes(type_)
        return validator(obj, type_args)

    return isinstance(obj, type_)


def get_origin(cls: typing.Any) -> typing.Any:
    """Examples:
    >>> get_origin(typing.List[float])
    >>> <class 'list'>
    """
    return typing.get_origin(cls) if hasattr(typing, 'get_origin') else cls.__origin__


def get_type_arguments(cls: typing.Any) -> typing.Tuple[typing.Any, ...]:
    """Examples:
    >>> get_type_arguments(int)
    >>> ()
    >>> get_type_arguments(typing.List)
    >>> ()
    >>> get_type_arguments(typing.List[int])
    >>> (<class 'int'>,)
    >>> get_type_arguments(typing.Callable[[int, float], str])
    >>> ([<class 'int'>, <class 'float'>], <class 'str'>)
    >>> get_type_arguments(typing.Callable[..., str])
    >>> (Ellipses, <class 'str'>)
    """
    if hasattr(typing, 'get_args'):
        # Python 3.8.0 throws index error here, for argument cls = typing.Callable (without type arguments)
        try:
            return __clean_type_arguments(args=typing.get_args(cls))
        except IndexError:
            return ()
    elif hasattr(cls, '__args__'):
        # return cls.__args__  # DOESNT WORK. So below is the modified (!) implementation of typing.get_args()

        res = __clean_type_arguments(args=cls.__args__)
        if res == ():
            return ()

        origin = get_origin(cls)
        if ((origin is typing.Callable) or (origin is collections.abc.Callable)) and res[0] is not Ellipsis:
            res = (list(res[:-1]), res[-1])
        return res
    else:
        return ()


def __clean_type_arguments(args: typing.Tuple[typing.Any, ...]) -> typing.Tuple[typing.Any, ...]:
    if args is None or args == () or any([isinstance(o, typing.TypeVar) for o in args]):
        return ()
    return args


def has_required_type_arguments(cls: typing.Any) -> bool:
    """Examples:
    >>> has_required_type_arguments(typing.List)
    >>> False
    >>> has_required_type_arguments(typing.List[int])
    >>> True
    >>> has_required_type_arguments(typing.Callable[[int, float], typing.Tuple[float, str]])
    >>> True
    """
    num_type_args = len(get_type_arguments(cls))
    requirements_exact = {
        'typing.Callable': 2,
        'typing.List': 1,
        'typing.Set': 1,
        'typing.FrozenSet': 1,
        'typing.Iterable': 1,
        'typing.Dict': 2,
        'typing.Optional': 2,  # because typing.get_args(typing.Optional[int]) returns (int, None)
    }
    requirements_min = {
        'typing.Tuple': 1,
        'typing.Union': 2,
    }
    base: str = get_base_class_as_str(cls=cls)

    if base in requirements_exact:
        return requirements_exact[base] == num_type_args
    elif base in requirements_min:
        return requirements_min[base] <= num_type_args
    else:
        return True


def get_base_class_as_str(cls: typing.Any) -> str:
    """Example:
    >>> get_base_class_as_str(typing.List[int])
    >>> 'typing.List'
    """
    return str(cls).split('[')[0]


if hasattr(typing, '_GenericAlias'):
    def _is_generic(cls):

        if isinstance(cls, typing._GenericAlias):
            return True

        if isinstance(cls, typing._SpecialForm):
            return cls not in {typing.Any}
        return False


    def _is_base_generic(cls) -> bool:
        # return len(typing.get_args(cls)) == 0

        if isinstance(cls, typing._GenericAlias):
            if cls.__origin__ in {typing.Generic, typing_protocol}:
                return False

            # inserted by Willi to make is compatible with Python versions 3.9 or later
            is_variadic_generic_alias = str(cls) in ['typing.Tuple', 'typing.Callable']
            if hasattr(typing, "_VariadicGenericAlias") and not \
                    (isinstance(cls, typing._VariadicGenericAlias) == is_variadic_generic_alias):
                raise ValueError(f'Assumption made during refactoring failed: {is_variadic_generic_alias}')

            if is_variadic_generic_alias:
                return True

            return len(cls.__parameters__) > 0

        if isinstance(cls, typing._SpecialForm):
            return cls._name in {'ClassVar', 'Union', 'Optional'}

        return False


    def _get_base_generic(cls):
        # subclasses of Generic will have their _name set to None, but
        # their __origin__ will point to the base generic
        if cls._name is None:
            return cls.__origin__
        else:
            return getattr(typing, cls._name)


    def _get_python_type(cls):
        """Like `python_type`, but only works with `typing` classes."""
        return cls.__origin__


    def _get_name(cls):
        return cls._name
else:
    def _is_generic(cls):
        if isinstance(cls, (typing.GenericMeta, typing._Union, typing._Optional, typing._ClassVar)):
            return True
        return False


    def _is_base_generic(cls):
        if isinstance(cls, (typing.GenericMeta, typing._Union)):
            return cls.__args__ in {None, ()}

        if isinstance(cls, typing._Optional):
            return True
        return False

    def _get_base_generic(cls):
        return cls.__origin__

    def _get_python_type(cls):
        """Like `python_type`, but only works with `typing` classes."""
        # Many classes actually reference their corresponding abstract base class from the abc module
        # instead of their builtin variant (i.e. typing.List references MutableSequence instead of list).
        # We're interested in the builtin class (if any), so we'll traverse the MRO and look for it there.
        for typ in cls.mro():
            if typ.__module__ == 'builtins' and typ is not object:
                return typ

    def _get_name(cls):
        try:
            return cls.__name__
        except AttributeError:
            return type(cls).__name__[1:]


def _get_subtypes(cls):
    subtypes = cls.__args__

    if get_base_generic(cls) is typing.Callable:
        if len(subtypes) != 2 or subtypes[0] is not ...:
            subtypes = (subtypes[:-1], subtypes[-1])
    return subtypes


def is_generic(cls):
    """
    Detects any kind of generic, for example `List` or `List[int]`. This includes "special" types like
    Union and Tuple - anything that's subscriptable, basically.
    """
    return _is_generic(cls)


def is_base_generic(cls: typing.Any) -> bool:
    """
    Detects generic base classes.

    Examples:
    >>> is_base_generic(int)  # int, float, str, bool
    >>> False
    >>> is_base_generic(typing.List)
    >>> True
    >>> is_base_generic(typing.Callable)
    >>> True
    >>> is_base_generic(typing.List[int])
    >>> False
    >>> is_base_generic(typing.Callable[[int], str])
    >>> False
    >>> is_base_generic(typing.List[typing.List[int]])
    >>> False
    """
    assert has_required_type_arguments(cls), \
        f'The type annotation "{cls}" misses some type arguments e.g. ' \
        f'"typing.Tuple[Any, ...]" or "typing.Callable[..., str]".'
    return _is_base_generic(cls)


def is_qualified_generic(cls):
    """
    Detects generics with arguments, for example `List[int]` (but not `List`)
    """
    return is_generic(cls) and not is_base_generic(cls)


def get_base_generic(cls):
    return _get_base_generic(cls)


def get_subtypes(cls):
    return _get_subtypes(cls)


def _instancecheck_iterable(iterable, type_args):
    type_ = type_args[0]
    return all(is_instance(val, type_) for val in iterable)


def _instancecheck_mapping(mapping, type_args):
    return _instancecheck_itemsview(mapping.items(), type_args)


def _instancecheck_itemsview(itemsview, type_args):
    key_type, value_type = type_args
    return all(is_instance(key, key_type) and is_instance(val, value_type) for key, val in itemsview)


def _instancecheck_tuple(tup, type_args) -> bool:
    """Examples:
    >>> _instancecheck_tuple(tup=(42.0, 43, 'hi', 'you'), type_args=(typing.Any, Ellipsis))  # = Tuple[Any, ...]
    >>> True
    """
    if Ellipsis in type_args:
        return all(is_instance(val, type_args[0]) for val in tup)

    if len(tup) != len(type_args):
        return False

    return all(is_instance(val, type_) for val, type_ in zip(tup, type_args))


_ORIGIN_TYPE_CHECKERS = {}
for class_path, check_func in {
    # iterables
    'typing.Container': _instancecheck_iterable,
    'typing.Collection': _instancecheck_iterable,
    'typing.AbstractSet': _instancecheck_iterable,
    'typing.MutableSet': _instancecheck_iterable,
    'typing.Sequence': _instancecheck_iterable,
    'typing.MutableSequence': _instancecheck_iterable,
    'typing.ByteString': _instancecheck_iterable,
    'typing.Deque': _instancecheck_iterable,
    'typing.List': _instancecheck_iterable,
    'typing.Set': _instancecheck_iterable,
    'typing.FrozenSet': _instancecheck_iterable,
    'typing.KeysView': _instancecheck_iterable,
    'typing.ValuesView': _instancecheck_iterable,
    'typing.AsyncIterable': _instancecheck_iterable,

    # mappings
    'typing.Mapping': _instancecheck_mapping,
    'typing.MutableMapping': _instancecheck_mapping,
    'typing.MappingView': _instancecheck_mapping,
    'typing.ItemsView': _instancecheck_itemsview,
    'typing.Dict': _instancecheck_mapping,
    'typing.DefaultDict': _instancecheck_mapping,
    'typing.Counter': _instancecheck_mapping,
    'typing.ChainMap': _instancecheck_mapping,

    # other
    'typing.Tuple': _instancecheck_tuple,
}.items():
    try:
        cls = eval(class_path)
    except AttributeError:
        continue

    _ORIGIN_TYPE_CHECKERS[cls] = check_func


def _instancecheck_callable(value, type_):
    if not callable(value):
        return False

    if is_base_generic(type_):
        return True

    param_types, ret_type = get_subtypes(type_)
    sig = inspect.signature(value)

    missing_annotations = []

    if param_types is not ...:
        if len(param_types) != len(sig.parameters):
            return False

        # if any of the existing annotations don't match the type, we'll return False.
        # Then, if any annotations are missing, we'll throw an exception.
        for param, expected_type in zip(sig.parameters.values(), param_types):
            param_type = param.annotation
            if param_type is inspect.Parameter.empty:
                missing_annotations.append(param)
                continue

            if not is_subtype(param_type, expected_type):
                return False

    if sig.return_annotation is inspect.Signature.empty:
        missing_annotations.append('return')
    else:
        if not is_subtype(sig.return_annotation, ret_type):
            return False

    assert not missing_annotations,\
        f'Parsing of type annotations failed. Maybe you are about to return a lambda expression. ' \
        f'Try returning an inner function instead. {missing_annotations}'
    return True


def _instancecheck_union(value, type_):
    types = get_subtypes(type_)
    return any(is_instance(value, typ) for typ in types)


_SPECIAL_INSTANCE_CHECKERS = {
    'Union': _instancecheck_union,
    'Callable': _instancecheck_callable,
    'Any': lambda v, t: True,
}


def is_subtype(sub_type, super_type):
    if not is_generic(sub_type):
        python_super = python_type(super_type)
        python_sub = python_type(sub_type)
        return issubclass(python_sub, python_super)

    # at this point we know `sub_type` is a generic
    python_sub = python_type(sub_type)
    python_super = python_type(super_type)
    if not issubclass(python_sub, python_super):
        return False

    # at this point we know that `sub_type`'s base type is a subtype of `super_type`'s base type.
    # If `super_type` isn't qualified, then there's nothing more to do.
    if not is_generic(super_type) or is_base_generic(super_type):
        return True

    # at this point we know that `super_type` is a qualified generic... so if `sub_type` isn't
    # qualified, it can't be a subtype.
    if is_base_generic(sub_type):
        return False

    # at this point we know that both types are qualified generics, so we just have to
    # compare their sub-types.
    sub_args = get_subtypes(sub_type)
    super_args = get_subtypes(super_type)
    return all(is_subtype(sub_arg, super_arg) for sub_arg, super_arg in zip(sub_args, super_args))


def python_type(annotation):
    """
    Given a type annotation or a class as input, returns the corresponding python class.

    Examples:
        >>> python_type(typing.Dict)
        <class 'dict'>
        >>> python_type(typing.List[int])
        <class 'list'>
        >>> python_type(int)
        <class 'int'>
        >>> python_type(typing.Any)
        <class 'object'>
    """
    if hasattr(annotation, 'mro'):
        mro = annotation.mro()
        if typing.Type in mro:
            return annotation.python_type
        elif annotation.__module__ == 'typing':
            return _get_python_type(annotation)
        else:
            return annotation
    elif annotation == typing.Any or annotation == Ellipsis:
        return object
    else:
        return _get_python_type(annotation)



from setuptools import find_packages, setup
from os import path

url = "https://github.com/LostInDarkMath/pedantic-python-decorators"
author = "Willi Sontopski"

# read the contents of the README file
this_directory = path.abspath(path.dirname(__file__))
with open(path.join(this_directory, 'README.md'), encoding='utf-8') as f:
    long_description = f.read()

setup(
    # important things
    name="pedantic",
    version="1.0.6",
    packages=find_packages(),
    # scripts = ["say_hello.py"]
    install_requires=['docstring_parser'],
    # package_data={
    #     # If any package contains *.txt or *.rst files, include them:
    #     "": ["*.txt", "*.rst"],
    #     # And include any *.msg files found in the "hello" package, too:
    #     "hello": ["*.msg"],
    # },

    # metadata to display on PyPI
    author=author,
    # author_email="me@example.com",
    license="Apache-2.0 License",
    maintainer=author,
    description="Some useful Python decorators for cleaner software development.",
    long_description=long_description,
    long_description_content_type='text/markdown',
    keywords="decorators tools helpers type-checking pedantic type annotations",
    url=url,
    project_urls={
        "Bug Tracker": f'{url}/issues',
        "Documentation": url,
        "Source Code": url,
    },

    # less important things
    include_package_data=False,
    zip_safe=True,
)
